// Copyright 2020 vrqq. (http://vrqq.org)

// This file is part of iOSAnywhere.

// iOSAnywhere is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// iOSAnywhere is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.

// You should have received a copy of the GNU Lesser General Public License
// along with iOSAnywhere.  If not, see <https://www.gnu.org/licenses/>.

#include <iostream>
#include <thread>
#include <string>
#include <cstring>

#include <pthread.h>
#include <unistd.h>

#include <libimobiledevice/libimobiledevice.h>
#include <libimobiledevice/lockdown.h>
#include <libimobiledevice/service.h>
using namespace std;

pthread_spinlock_t imgloader_lock;
pthread_mutex_t locate_lock;
pthread_mutex_t searchdev_lock;
FILE *imgloader = nullptr, FILE_PLACEHOLDER;

char udid[256];
void searchdev() {
	if (pthread_mutex_trylock(&searchdev_lock) != 0)
		return ;
	char first_udid[256];
	memset(first_udid, 0, sizeof(first_udid));
	char** dev_list = NULL;
	int device_size;
	if (idevice_get_device_list(&dev_list, &device_size) < 0) {
		fprintf(stderr, "ERROR: Unable to retrieve device list!\n");
		memset(udid, 0, sizeof(udid));
		pthread_mutex_unlock(&searchdev_lock);
		return ;
	}
	if (dev_list[0] != nullptr)
		memcpy(udid, dev_list[0], strlen(dev_list[0]));
	idevice_device_list_free(dev_list);
	pthread_mutex_unlock(&searchdev_lock);
}


struct Screen {
	const static int MAXLOGSIZE = 50*1024;
	char buffer[MAXLOGSIZE];
	int submitted=0, cursor=0, size=0;
	
	std::string output() {
		// return to_string(size);
		return std::string(buffer, size);
	}

	void append(char c) {
		if (c == '\r')
			cursor = 0;
		else if (c == '\b')
			cursor--;
		else {
			if (submitted + cursor >= MAXLOGSIZE)
				return ;
			buffer[submitted + cursor++] = c;
			if (submitted + cursor > size)
				size = submitted + cursor;
			if (c == '\n')
				submitted = size, cursor = 0;
		}
	}

	void clear() {
		submitted = cursor = size = 0;
	}
}screen;


//thread unsafe.
pthread_t thread_logcache;
void* funcLogCache(void *) {
	cout<<"* Start Logcapture."<<endl;
	// char ch;
	// while(fread(&ch, sizeof(char), 1, imgloader) > 0)
	int ch;
	while((ch = fgetc(imgloader)) != EOF)
		screen.append((char)ch);

	pthread_spin_lock(&imgloader_lock);
	pclose(imgloader);
	imgloader = nullptr;
	pthread_spin_unlock(&imgloader_lock);
	cout<<"* Stop Logcapture."<<endl<<endl;
	return 0;
}


//web interface - sync
std::string getInjectLog() {
	const std::string result = screen.output();
	pthread_spin_lock(&imgloader_lock);
	if (imgloader == nullptr) // return "" for next request log.
		screen.clear();
	pthread_spin_unlock(&imgloader_lock);
	return result;
}

// web interface - async
void inject(std::string ver) {
	pthread_spin_lock(&imgloader_lock);
	if (imgloader != nullptr) {
		pthread_spin_unlock(&imgloader_lock);
		return ;
	}
	imgloader = &FILE_PLACEHOLDER;
	pthread_spin_unlock(&imgloader_lock);
	cout<<"Development Image injector "<<ver<<endl;
	
	std::string img = "driver/"+ver+"/DeveloperDiskImage.dmg";
	std::string sign = "driver/"+ver+"/DeveloperDiskImage.dmg.signature";
	std::string cmd = "ideviceimagemounter " + img + " " + sign + " 2>&1";
	FILE *tmp = popen(cmd.c_str(), "r");
	// FILE *tmp = popen("tput cols", "r");
	// FILE *tmp = popen("ping 127.0.0.1 -c 5 2>&1", "r");
	// FILE *tmp = popen("wget http://kd.269.net/200.zip 2>&1", "r");
	if (!tmp) {
		pthread_spin_lock(&imgloader_lock);
		imgloader = nullptr;
		pthread_spin_unlock(&imgloader_lock);
		return ;
	}
	imgloader = tmp;
	pthread_create(&thread_logcache, NULL, funcLogCache, NULL);

}


//web interface - sync
std::string locate(std::string lat1, std::string lng2) {
	if (pthread_mutex_trylock(&locate_lock) != 0)
		return "busy";
	searchdev();
	cout<<"Device "<<udid<<endl;
	cout<<"\tSet to latitude="<<lat1<<" longitude="<<lng2<<endl<<endl;
	idevice_t device = nullptr;
	lockdownd_client_t lckd = nullptr;
	lockdownd_service_descriptor_t service = nullptr;
	service_client_t client = nullptr;
	if (IDEVICE_E_SUCCESS != idevice_new(&device, udid)) {
		if (udid)
			std::cout << "\tNo device found with udid " << udid << ", is it plugged in?" << std::endl;
		else
			std::cout << "\tNo device found, is it plugged in?" << std::endl;
		pthread_mutex_unlock(&locate_lock);
		return "No device found, is it plugged in?";
	}

	lockdownd_error_t ldret;
	if (LOCKDOWN_E_SUCCESS != (ldret = lockdownd_client_new_with_handshake(device, &lckd, NULL))) {
		idevice_free(device);
		std::cout << "\tERROR: Could not connect to lockdownd, error code "<< ldret << std::endl;
		pthread_mutex_unlock(&locate_lock);
		return "ERROR: Could not connect to lockdownd, error code " + to_string(ldret);
	}
	if (LOCKDOWN_E_SUCCESS != (ldret = lockdownd_start_service(lckd, "com.apple.dt.simulatelocation", &service))) {
		pthread_mutex_unlock(&locate_lock);
		std::cout<<"\tERROR: lockdownd_start_service failure, error code "<<to_string(ldret);
		return "ERROR: lockdownd_start_service failure, error code " + to_string(ldret);
	}
	lockdownd_client_free(lckd);

	service_error_t svret;
	if (SERVICE_E_SUCCESS != (svret = service_client_new(device, service, &client))) {
		pthread_mutex_unlock(&locate_lock);
		std::cout<<"\tERROR: service_client_new failure, error code "<<to_string(ldret);
		return "ERROR: service_client_new failure, error code " + to_string(svret);
	}

	// start sending lng/lat
	char datasize[4] = {0, 0, 0, 0};
	auto mksize = [&](size_t sz) {
		union {
			uint32_t uintval;
			char charval[4];
		}test, cnv;
		cnv.uintval = sz;
		test.uintval = 0x1;
		if (test.charval[0] == 0x1) {
			datasize[0] = cnv.charval[3]; datasize[1] = cnv.charval[2];
			datasize[2] = cnv.charval[1]; datasize[3] = cnv.charval[0];
		}
		else {
			datasize[0] = cnv.charval[0]; datasize[1] = cnv.charval[1];
			datasize[2] = cnv.charval[2]; datasize[3] = cnv.charval[3];
		}
	};
	size_t sent;
	service_send(client, datasize, sizeof(datasize), &sent);

	mksize(lat1.size());
	// cout<<(int)datasize[0]<<" "<<(int)datasize[1]<<" "<<(int)datasize[2]<<" "<<(int)datasize[3]<<endl;
	service_send(client, datasize, sizeof(datasize), &sent);
	service_send(client, lat1.c_str(), lat1.size(), &sent);
	mksize(lng2.size());
	service_send(client, datasize, sizeof(datasize), &sent);
	service_send(client, lng2.c_str(), lng2.size(), &sent);
	service_client_free(client);

	if (service)
		lockdownd_service_descriptor_free(service);
	idevice_free(device);
	pthread_mutex_unlock(&locate_lock);
	return "Updated to latitude=" + lat1 + ", longitude=" + lng2;
}

//web interface - sync
std::string getUdid() {
	searchdev();
	return std::string(udid, min(strlen(udid), (size_t)256));
}

void init2() {
	//PHREAD_PROCESS_PRIVATE or PTHREAD_PROCESS_SHARED
	pthread_spin_init(&imgloader_lock, PTHREAD_PROCESS_PRIVATE);
	pthread_mutex_init(&locate_lock, 0);
	pthread_mutex_init(&searchdev_lock, 0);
}

#ifndef DIST
int main() {
	//PHREAD_PROCESS_PRIVATE or PTHREAD_PROCESS_SHARED
	pthread_spin_init(&imgloader_lock, PTHREAD_PROCESS_PRIVATE);
	pthread_mutex_init(&locate_lock, 0);

	// inject("13.2");
	// for(int i=0; i<5; i++) {
	// 	cout<<"[LOG]"<<getInjectLog()<<endl<<endl;
	// 	// usleep(500*1000);
	// 	sleep(1);
	// };
	cout<<locate("39.0302", "127.8901")<<endl;

	pthread_mutex_destroy(&locate_lock);
	return 0;
}
#endif